from PySide2.QtCore import *
from PySide2.QtWidgets import *
from PySide2.QtGui import *
import sqlite3
import json
import os

from cutevariant.gui.wizard.projetwizard import ProjetWizard
from cutevariant.gui.settings import SettingsWidget
from cutevariant.gui.viewquerywidget import ViewQueryWidget
from cutevariant.gui.columnquerywidget import ColumnQueryWidget
from cutevariant.gui.filterquerywidget import FilterQueryWidget
from cutevariant.gui.selectionquerywidget import SelectionQueryWidget
from cutevariant.gui.hpoquerywidget import HpoQueryWidget
from cutevariant.gui.vqleditor import VqlEditor

from cutevariant.gui.queryrouter import QueryRouter

from cutevariant.core.importer import import_file
from cutevariant.core import Query

from cutevariant.gui.plugins.infovariantplugin import InfoVariantPlugin


class MainWindow(QMainWindow):
    def __init__(self, parent=None):
        super(MainWindow, self).__init__()
        self.toolbar = self.addToolBar("maintoolbar")
        self.conn = None
        self.view_widgets = []
        self.column_widget = ColumnQueryWidget()
        self.filter_widget = FilterQueryWidget()
        self.selection_widget = SelectionQueryWidget()
        self.tab_view = QTabWidget()
        self.editor = VqlEditor()

        # Setup Actions
        self.setupActions()

        # Init router
        self.router = QueryRouter()
        # self.router.addWidget(self.view_widget)
        self.router.addWidget(self.column_widget)
        self.router.addWidget(self.filter_widget)
        self.router.addWidget(self.selection_widget)
        self.router.addWidget(self.editor)

        vsplit = QSplitter(Qt.Vertical)
        vsplit.addWidget(self.tab_view)
        vsplit.addWidget(self.editor)

        self.setCentralWidget(vsplit)

        # self.test = InfoVariantPlugin()
        # self.test2 = GenomeView()

        #  Init panel
        self.addPanel(self.column_widget)
        self.addPanel(self.filter_widget)
        self.addPanel(self.selection_widget)
        # self.addPanel(self.test)
        # self.addPanel(self.test2)

        # self.addPanel(HpoQueryWidget())

        self.addView()

        # self.currentView().variant_clicked.connect(self.test.set_variant)
        # self.currentView().variant_clicked.connect(self.test2.set_variant)

        #  window geometry
        self.resize(600, 400)

        # self.import_vcf("/home/schutz/Dev/CuteVariant-python/exemples/test.snp.eff.vcf")

        # self.open("/home/schutz/Dev/CuteVariant-python/exemples/test.snpeff.vcf.db")

        self.setGeometry(qApp.desktop().rect().adjusted(100, 100, -100, -100))

    def import_vcf(self, filename):  #  Temporary .. will be removed
        db_filename = filename + ".db"

        if os.path.exists(db_filename):
            os.remove(db_filename)

        self.conn = sqlite3.connect(db_filename)
        import_file(self.conn, filename)

        self.conn.close()
        self.open(db_filename)

    def open(self, db_filename):

        if not os.path.exists(db_filename):
            QMessageBox.warning(self, "error", "file doesn't exists")
            return

        self.conn = sqlite3.connect(db_filename)
        query = Query(self.conn)
        # query.filter = json.loads(
        #     """{"AND" : [{"field":"pos", "operator":">", "value":"880000"} ]}"""
        # )

        # query.create_selection("mytest")
        # query.filter = None

        self.router.setQuery(query)

    def addPanel(self, widget, area=Qt.LeftDockWidgetArea):
        dock = QDockWidget()
        dock.setWindowTitle(widget.windowTitle())
        dock.setWidget(widget)
        self.addDockWidget(area, dock)
        self.view_menu.addAction(dock.toggleViewAction())

    def setupActions(self):
        # menu bar
        self.file_menu = self.menuBar().addMenu("&File")
        self.file_menu.addAction(
            "&New project", self, SLOT("new_project()"), QKeySequence.New
        )
        self.file_menu.addAction(
            "&Open project ...", self, SLOT("open_project()"), QKeySequence.Open
        )
        self.file_menu.addSeparator()
        self.file_menu.addAction("Settings ...", self, SLOT("show_settings()"))

        self.view_menu = self.menuBar().addMenu("&View")

        self.help_menu = self.menuBar().addMenu("Help")
        self.help_menu.addAction("About Qt", qApp, SLOT("aboutQt()"))

        # Tool bar

        save_query_action = self.toolbar.addAction("save query")
        save_query_action.triggered.connect(self.selection_widget.save_current_query)

    def addView(self):
        widget = ViewQueryWidget()
        self.view_widgets.append(widget)
        self.tab_view.addTab(widget, widget.windowTitle())
        self.router.addWidget(widget)

    def currentView(self):
        index = self.tab_view.currentIndex()
        if index == -1:
            return None
        return self.view_widgets[index]

    @Slot()
    def new_project(self):
        wizard = ProjetWizard()
        if wizard.exec_():
            db_filename = (
                wizard.field("project_path")
                + QDir.separator()
                + wizard.field("project_name")
                + ".db"
            )
            self.open(db_filename)

    @Slot()
    def open_project(self):
        filename = QFileDialog.getOpenFileName(
            self, "Open project", "Cutevariant project (*.db)"
        )[0]
        if filename is not None:
            self.open(filename)

    @Slot()
    def show_settings(self):
        widget = SettingsWidget()
        widget.exec_()



from PySide2.QtCore import *
from PySide2.QtWidgets import *
from PySide2.QtGui import *
from cutevariant.gui.variantplugin import VariantPlugin


class InfoVariantPlugin(VariantPlugin):
    def __init__(self):
        super().__init__()

        self.view = QTreeWidget()
        self.view.setColumnCount(2)
        v_layout = QVBoxLayout()
        v_layout.setContentsMargins(0, 0, 0, 0)
        v_layout.addWidget(self.view)
        self.setLayout(v_layout)

    def set_variant(self, variant):
        self.view.clear()

        # print(variant)
        for key, val in variant.items():
            item = QTreeWidgetItem()
            item.setText(0, str(key))
            item.setText(1, str(val))
            # item.setFlags(Qt.ItemIsSelectable|Qt.ItemIsEnabled)

            self.view.addTopLevelItem(item)



from PySide2.QtWidgets import *
from PySide2.QtCore import *
from PySide2.QtGui import *

from cutevariant.gui.abstractquerywidget import AbstractQueryWidget
from cutevariant.core import Query


class QueryRouter(QObject):
    """ 
	This class redirect query between widgets. 
	If one of the widget emit a 'changed' signal then all belongs widgets change their query except the sender.
	"""

    def __init__(self):
        super().__init__()
        self.widgets = []
        self.query = None

    def addWidget(self, widget: AbstractQueryWidget):
        """
        Add a widget into the router 

        :param widget: a query widget 
        """
        self.widgets.append(widget)
        widget.changed.connect(self.widgetChanged)

    def setQuery(self, query: Query):
        """
        Update query for all widgets 

        :param query: a Query object 
        """
        self.query = query
        for widget in self.widgets:
            widget.setQuery(self.query)

    def getQuery(self):
        """
        :return: Return current query 
        """
        return self.query

    def widgetChanged(self):
        """ 
        this method is trigger from one widget 
        """

        #  Get the wiget which send the signal changed
        sender_widget = self.sender()

        if sender_widget is not None:
            #  update query from sender widget
            self.query = sender_widget.getQuery()

            print(self.query)
            print(self.query.sql())

            #  change query for all widget except sender
            for widget in self.widgets:
                if widget != sender_widget:
                    widget.setQuery(self.query)



from abc import ABC, abstractclassmethod


class AbstractReader(ABC):
    """ 
    This is the base class for all Reader required to import variants into database.


    """

    def __init__(self, device):
        super(AbstractReader, self).__init__()
        self.device = device
        self.file_size = 0

    @abstractclassmethod
    def parse_variants(self):
        """
        This abstract method must return variants as a list of dictionnary. 
        
        Minimum output:
        ===============
        [
        {"chr": "chr3","pos": 3244,"ref": "A","alt":"C"},
        {"chr": "chr4","pos": 3244,"ref": "A","alt":"C"},
        {"chr": "chr5","pos": 3244,"ref": "A","alt":"C"}
        ]

        Full output:
        ============
        [
        {"chr": "chr3",
        "pos": 3244,
        "ref": "A",
        "alt":"C", 
        "field_n": "value_n",
        "annotations": [ 
            {"gene": "GJB2", "transcripts": "NM_00232.1", "field_n": "value_n"},
            {"gene": "GJB2", "transcripts": "NM_00232.2", "field_n": "value_n"}
            ],
        "samples": [
            {"name":"boby", "gt": 1, "field_n":"value_n"},
            {"name":"kevin", "gt": 1, "field_n":"value_n"}
            ]}
        ]
        """
        raise NotImplemented()

    @abstractclassmethod
    def parse_fields(self):
        """
        This abstract methods must return fields description defined from parse_variant output.
        You must define sqlite type for each field (text, integer, bool)

        Full output:
        ==============
        [
        {"name": "chr", "type": "text", "category": "variant", "description": "description"}, 
        {"name": "pos", "type": "text", "category": "variant", "description": "description"}, 
        {"name": "ref", "type": "text", "category": "variant", "description": "description"}, 
        {"name": "alt", "type": "text", "category": "variant", "description": "description"}, 
        {"name": "field_n", "type": "text", "category": "variant", "description": "description"},
        {"name": "name", "type": "text", "category": "annotations", "samples": "description"},
        {"name": "gt", "type": "text", "category": "annotations", "samples": "description"}
        ]
       """
        raise NotImplemented()

    def get_variants_count(self):
        """ 
        Return variant count. You can overload this method to make it faster
        """
        count = 0
        for v in self.get_variants():
            count += 1
        return count

    def get_samples(self):
        return []

    def get_fields(self):
        """decorator for get_fields"""
        yield from self.parse_fields()
        yield {
            "name": "description",
            "type": "text",
            "category": "extra",
            "description": "description of variant",
        }
        yield {
            "name": "favoris",
            "type": "bool",
            "category": "extra",
            "description": "is favoris",
            "default": False,
        }

    def get_variants(self):
        """decorator for get_fields"""
        yield from self.parse_variants()



import re
from . import sql
from . import vql


class Query:
    """
    This class is intended to build sql query according parameters
    self.columns : columns from variant table
    self.filter : filter filter as raw text
    self.selection : name of the variant set. Use "all" to select all variants
    """

    def __init__(
        self, conn, columns=["chr", "pos", "ref", "alt"], filter=None, selection="all"
    ):
        self.conn = conn
        self.columns = columns
        self.filter = filter
        self.selection = selection
        self.order_by = None
        self.order_desc = True

        ##-----------------------------------------------------------------------------------------------------------

    def sample_from_expression(self, expression):
        # extract <sample> from <gt("sample")>
        regexp = r"gt(.*).gt"
        match = re.search(regexp, expression)
        if match:
            return match.group(1)
        else:
            return None

        ##-----------------------------------------------------------------------------------------------------------

    def detect_samples(self):
        """ detect if query need sample join by looking genotype expression : genotype("boby").gt and return samples """
        # extract sample name from select and filter clause
        samples_detected = []
        combine_clause = self.columns

        for col in combine_clause:
            sample = self.sample_from_expression(col)
            if sample is not None:
                samples_detected.append(sample)

        if len(samples_detected) == 0:
            return {}
        # Look in DB if sample exists and returns {sample:id} dictionnary
        in_clause = ",".join([f"'{sample}'" for sample in samples_detected])

        return dict(
            self.conn.execute(
                f"SELECT name, rowid FROM samples WHERE name IN ({in_clause})"
            ).fetchall()
        )

        ##-----------------------------------------------------------------------------------------------------------

    def sql(self, limit=0, offset=0):
        """ build query depending class parameter """

        #  Detect if join sample is required ...
        sample_ids = self.detect_samples()

        if len(self.columns) == 0:
            self.columns = ["chr", "pos", "ref", "alt"]

        #  Replace columns gt(sacha) by sv4.gt ( where 4 is the sample id for outer join)
        sql_columns = []
        sql_columns.append("variants.rowid")
        for col in self.columns:
            sample = self.sample_from_expression(col)
            if sample is not None:
                sql_columns.append(f"gt{sample}.gt")
            else:
                sql_columns.append(col)

        query = f"SELECT {','.join(sql_columns)} "

        # Add Select clause

        if self.selection == "all":
            query += f"FROM variants"
        else:
            #  manage jointure with selection

            query += f"""
            FROM variants
            INNER JOIN selection_has_variant sv ON sv.variant_id = variants.rowid
            INNER JOIN selections s ON s.rowid = sv.selection_id AND s.name = '{self.selection}'
            """

        if len(sample_ids):
            for sample, i in sample_ids.items():
                query += f" LEFT JOIN sample_has_variant gt{sample} ON gt{sample}.variant_id = variants.rowid AND gt{sample}.sample_id = {i} "

                # add filter clause
        if self.filter:
            query += " WHERE " + self.filter_to_sql(self.filter)
            #  add limit and offset

        if self.order_by is not None:
            direction = "DESC" if self.order_desc is True else "ASC"
            query += f" ORDER BY {self.order_by} {direction}"

        if limit > 0:
            query += f" LIMIT {limit} OFFSET {offset}"

        print(query)
        return query

        ##-----------------------------------------------------------------------------------------------------------

    def rows(self, limit=0, offset=0):
        """ return query results as list by record """
        print(self.sql(limit, offset))
        yield from self.conn.execute(self.sql(limit, offset))

        ##-----------------------------------------------------------------------------------------------------------

    def items(self, limit=0, offset=0):
        """ return query results as dict by record """
        for value in self.conn.execute(self.sql(limit, offset)):
            item = {}
            for index, col in enumerate(["rowid"] + self.columns):
                item[col] = value[index]
            yield item

        ##-----------------------------------------------------------------------------------------------------------

    def filter_to_sql(self, node: dict) -> str:

        if dict is None:
            return str()

        # function to detect if node is a Condition node (AND/OR) OR a field node {name,operator, value}
        is_field = lambda x: True if len(x) == 3 else False

        if is_field(node):
            # change value
            value = node["value"]
            operator = node["operator"]
            field = node["field"]

            if (
                type(value) == str
            ):  # Add quote for string .. Need to change in the future and use sqlite binding value
                value = "'" + str(value) + "'"
            else:
                value = str(value)

            #  change columns name for sample join
            sample = self.sample_from_expression(field)
            if sample:
                field = f"gt{sample}.gt"

            return field + operator + value

        else:
            logic = list(node.keys())[0]
            out = []
            for child in node[logic]:
                out.append(self.filter_to_sql(child))

            return "(" + f" {logic} ".join(out) + ")"

        ##-----------------------------------------------------------------------------------------------------------

    def samples(self):
        return self.detect_samples().keys()

        ##-----------------------------------------------------------------------------------------------------------

    def create_selection(self, name):
        sql.create_selection_from_sql(self.conn, name, self.sql())

        ##-----------------------------------------------------------------------------------------------------------

    def count(self):
        """ return total row number """
        #  TODO : need to cache this method because it can take time to compute with large dataset
        return self.conn.execute(
            f"SELECT COUNT(*) as count FROM ({self.sql()})"
        ).fetchone()[0]

        ##-----------------------------------------------------------------------------------------------------------

    def __repr__(self):
        return f"""
        columns : {self.columns}
        filter: {self.filter}
        selection: {self.selection}
        """

        ##-----------------------------------------------------------------------------------------------------------

    def from_vql(self, raw: str):
        model = vql.model_from_string(raw)
        self.columns = model["select"]
        self.selection = model["from"]
        self.filter = model.get("where")  # None if no filter
        # TODO: USING clause missing

        print("from vql", model)

        ##-----------------------------------------------------------------------------------------------------------

    def to_vql(self) -> str:
        base = f"SELECT {','.join(self.columns)} FROM {self.selection}"
        where = ""
        if self.filter:
            where = f" WHERE {self.filter_to_sql(self.filter)}"
        return base + where + ";"

        ##-----------------------------------------------------------------------------------------------------------

    def check(self):
        """ Return True if query is valid """
        return True



import os
import csv
import sqlite3
from .readerfactory import create_reader
from .sql import *


def async_import_file(conn, filename, project={}):
    """
    Import filename into sqlite connection

    :param conn: sqlite connection
    :param filenaame: variant filename 

    :return: yield progression and message

    """

    with create_reader(filename) as reader:

        #  Create projects
        create_project(
            conn, name=project.get("name"), reference=project.get("reference")
        )

        yield 0, "create table shema"
        #  Create table fields
        create_table_fields(conn)

        # Create table samples
        create_table_samples(conn)

        #  Create variants tables
        create_table_variants(conn, reader.get_fields())

        #  Create selection
        create_table_selections(conn)

        yield 0, "insert samples"

        #  insert samples
        for sample in reader.get_samples():
            insert_sample(conn, name=sample)

        # Insert fields
        yield 0, "insert fields"
        insert_many_fields(conn, reader.get_fields())

        yield 0, "count variants..."
        total_variant = reader.get_variants_count()
        yield from insert_many_variants(conn, reader.get_variants(), total_variant)

    # # Create default selection
    # session.add(Selection(name="all", description="all variant", count = variant_count))
    # session.add(Selection(name="favoris", description="favoris", count = 0))

    # session.commit()


def import_file(conn, filename):
    for progress, message in async_import_file(conn, filename):
        #  don't show message
        pass



from .reader import *
import os
import struct
from contextlib import contextmanager
import binascii
import pathlib


def is_gz_file(filepath):
    with open(filepath, "rb") as test_f:
        return binascii.hexlify(test_f.read(2)) == b"1f8b"


def getuncompressedsize(filename):
    with open(filename, "rb") as f:
        f.seek(-4, 2)
        return struct.unpack("I", f.read(4))[0]


@contextmanager
def create_reader(filename):

    path = pathlib.Path(filename)

    print("PATH suffix", path.suffixes, is_gz_file(filename))

    if ".vcf" in path.suffixes and ".gz" in path.suffixes:
        device = open(filename, "rb")
        reader = VcfReader(device)
        reader.file_size = getuncompressedsize(filename)
        yield reader
        device.close()
        return

    if ".vcf" in path.suffixes:
        device = open(filename, "r")
        reader = VcfReader(device)
        reader.file_size = os.path.getsize(filename)
        yield reader
        device.close()
        return

    if ".csv" in path.suffixes:
        device = open(filename, "r")
        reader = CsvReader(device)
        reader.file_size = os.path.getsize(filename)
        yield reader
        device.close()
        return


# class ReaderFactory(object):
#     """
# 	Create reader depending file type
# 	"""

#     def __init__(self):
#         pass

#     @staticmethod
#     def create_reader(filename):

#         if not os.path.isfile(filename):
#             raise FileExistsError()

#         extension = os.path.splitext(filename)[1]

#         if extension == ".csv":
#             print("create csv reader")
#             return CsvReader(open(filename, "r"))

#         if extension == ".vcf":
#             print("create vcf reader")
#             return VcfReader(open(filename, "r"))

#         if extension == ".vcf":
#             print("create vcf reader")
#             return VcfReader(open(filename, "r"))

#         return None



from PySide2.QtWidgets import *
from PySide2.QtCore import *


class BaseWidget(QTabWidget):
    def __init__(self):
        super().__init__()

    def save(self):
        raise NotImplemented()

    def load(self):
        raise NotImplemented()


class GeneralSettingsWidget(BaseWidget):
    def __init__(self):
        super().__init__()
        self.setWindowTitle("General")

    def save(self):
        pass

    def load(self):
        pass


class PluginsSettingsWidget(BaseWidget):
    def __init__(self):
        super().__init__()
        self.setWindowTitle("Plugins")

    def save(self):
        pass

    def load(self):
        pass


class SettingsWidget(QDialog):
    def __init__(self):
        super().__init__()
        self.widgets = []

        self.list_widget = QListWidget()
        self.stack_widget = QStackedWidget()
        self.button_box = QDialogButtonBox(
            QDialogButtonBox.SaveAll | QDialogButtonBox.Cancel | QDialogButtonBox.Reset
        )

        self.list_widget.setFixedWidth(200)
        self.list_widget.setIconSize(QSize(32, 32))

        h_layout = QHBoxLayout()
        h_layout.addWidget(self.list_widget)
        h_layout.addWidget(self.stack_widget)

        v_layout = QVBoxLayout()
        v_layout.addLayout(h_layout)
        v_layout.addWidget(self.button_box)
        self.setLayout(v_layout)

        self.addPanel(GeneralSettingsWidget())
        self.addPanel(PluginsSettingsWidget())

        self.resize(800, 400)

        self.button_box.button(QDialogButtonBox.SaveAll).clicked.connect(self.save_all)
        self.button_box.button(QDialogButtonBox.Reset).clicked.connect(self.load_all)
        self.button_box.button(QDialogButtonBox.Cancel).clicked.connect(self.close)

    def addPanel(self, widget: BaseWidget):
        self.widgets.append(widget)
        self.list_widget.addItem(
            QListWidgetItem(widget.windowIcon(), widget.windowTitle())
        )
        self.stack_widget.addWidget(widget)

    def save_all(self):
        for widget in self.widgets:
            widget.save()

    def load_all(self):
        for widget in self.widgets:
            widget.load()



"""Proof of concept for the VQL DSL.

See test_vql.py for usage and features.

"""


import textx
import operator
import itertools
from pprint import pprint
from pkg_resources import resource_string


CLAUSES = ("select", "from", "where", "using")
OPERATOR_PRECEDENCE = {
    op: (precedence, subprecedence)
    for precedence, ops in enumerate(
        map(
            str.split,
            """
    ||
    *    /    %
    +    -
    <<   >>   &    |
    <    <=   >    >=
    =    ==   !=   <>   IS   IS NOT   IN   LIKE   GLOB   MATCH   REGEXP
    AND  XOR
    OR
""".splitlines(),
        )
    )
    for subprecedence, op in enumerate(ops)
}


def precedence(one: operator, two: operator) -> operator:
    return one if OPERATOR_PRECEDENCE[one] >= OPERATOR_PRECEDENCE[two] else two


OPERATOR_FROM_LEXEM = {"==": "=", "=/=": "!=", "and": "AND", "or": "OR", "xor": "XOR"}


def model_class(name: str, bases: tuple, attrs: dict) -> type:
    """Metaclass to automatically build the __init__ to get the properties,
    and register the class for metamodel
    """
    if "__init__" not in attrs:

        def __init__(self, *args, **kwargs):
            for field, value in kwargs.items():
                setattr(self, field, value)

        attrs["__init__"] = __init__
    cls = type(name, bases, attrs)
    model_class.classes.append(cls)
    return cls


model_class.classes = []


# classes used to build the raw model
class RawCondition(metaclass=model_class):
    @property
    def value(self):
        return {
            "field": self.id.id,
            "operator": OPERATOR_FROM_LEXEM.get(self.op.op, self.op.op),
            "value": self.val if isinstance(self.val, (str, int)) else self.val.id,
        }


class ParenExpr(metaclass=model_class):
    @property
    def value(self):
        return self.expression.value


class BaseExpr(metaclass=model_class):
    @property
    def value(self):
        # get the infix tree describing the expression
        #  NB: expressions are nested instead of parenthesed
        stack = [self.left.value]
        for operation in self.operations:
            stack.extend(operation.value)
        return stack


class Operation(metaclass=model_class):
    @property
    def value(self):
        return (self.op.value, self.remaining.value)


class BoolOperator(metaclass=model_class):
    @property
    def value(self):
        return OPERATOR_FROM_LEXEM.get(self.op, self.op)


class Tuple(metaclass=model_class):
    @property
    def id(self):
        return "({})".format(", ".join(item.id for item in self.items))


METAMODEL = textx.metamodel_from_str(
    resource_string(__name__, "vql.tx").decode(),  # grammar extraction from vql.tx
    classes=model_class.classes,
    debug=False,
)


class VQLSyntaxError(ValueError):
    pass


def model_from_string(raw_vql: str) -> dict:
    try:
        raw_model = METAMODEL.model_from_str(raw_vql)
    except textx.exceptions.TextXSyntaxError as err:
        raise VQLSyntaxError(*error_message_from_err(err, raw_vql))
    model = {}
    for clause in CLAUSES:
        value = getattr(raw_model, clause)
        if value is not None:
            model[clause] = globals()[f"compile_{clause}_from_raw_model"](value)
    print("MODEL:", model)
    return model


def compile_select_from_raw_model(raw_model) -> dict or tuple:
    return tuple(column.id for column in raw_model.columns)


def compile_from_from_raw_model(raw_model) -> dict or tuple:
    return raw_model.source.id


def compile_where_from_raw_model(raw_model) -> dict or tuple:
    expr = raw_model.expression
    # print('EXPRESSION:', expr.value)
    tree = dicttree_from_infix_nested_stack(expr.value)
    # print('TREE:', tree)
    return tree


def compile_using_from_raw_model(raw_model) -> dict or tuple:
    return (raw_model.filename.id,)


def error_message_from_err(
    err: textx.exceptions.TextXSyntaxError, raw_vql: str
) -> (str, int):
    """Return human-readable information and index in raw_sql query
    about the given exception"""
    # print(err)
    # print(dir(err))
    # print(err.args)
    # print(err.err_type)
    # print(err.line, err.col)
    # print(err.message)
    # print(err.filename)
    # print(err.expected_rules)
    if "'SELECT'" in err.message:  # was awaiting for a SELECT clause
        return "no SELECT clause", -1
    if err.message.endswith("=> 's,ref FROM*'."):
        return "empty 'FROM' clause", err.col
    if (
        ",*," in err.message
        and len(err.expected_rules) == 1
        and type(err.expected_rules[0]).__name__ == "RegExMatch"
    ):
        return "invalid empty identifier in SELECT clause", err.col
    if "Expected INT " in err.message and len(err.expected_rules) == 3:
        return "invalid value in WHERE clause", err.col
    if "Expected '==|>=|<=|!=" in err.message and len(err.expected_rules) == 1:
        return "invalid operator in WHERE clause", err.col

    raise err  # error not handled. Just raise it


def dicttree_from_infix_nested_stack(stack: list) -> dict:
    "Return the dict tree computed from the infix nested stack yielded by metamodel"
    # Algorithm:
    #   expand nested structure, place parens
    #   build the postfix representation
    #   build tree as dict from postfix
    def expand(obj: list or tuple or ...) -> list:
        if isinstance(obj, (list, tuple)):
            return ("(", *itertools.chain(*map(expand, obj)), ")")
        return (obj,)

    def as_postfix(tokens: list) -> iter:
        "Yield the output describing the postfix notation of given list"
        operators = []  # stack of operator
        for token in tokens:
            print("TOKEN:", type(token), token)
            if isinstance(token, dict):  # it's a condition, like a=3
                yield token
            elif token in OPERATOR_PRECEDENCE:  # it's an operator
                while (
                    operators
                    and operators[-1] != "("
                    and precedence(token, operators[-1])
                ):
                    yield operators.pop()
                operators.append(token)
            elif token == "(":
                operators.append(token)
            elif token == ")":
                while operators and operators[-1] != "(":
                    yield operators.pop()
                if operators[-1] == "(":
                    operators.pop()
            else:
                raise VQLSyntaxError(f"Unexpected token: {token}")
        yield from operators

    def as_tree(postfix: list) -> dict:
        "Return the dict describing the n-ary syntax tree"
        tree = {}
        operandes = []
        last_operator = None
        for token in postfix:
            if isinstance(token, dict):  # it's a condition, like a=3
                operandes.append(token)
            elif token in OPERATOR_PRECEDENCE:  # it's an operator
                assert len(operandes) >= 2
                right, left = operandes.pop(), operandes.pop()
                if last_operator == token:  # let's merge everything
                    assert isinstance(left, dict)
                    assert len(left) == 1
                    assert isinstance(left[last_operator], list)
                    left[last_operator].append(right)
                    operandes.append(left)
                else:
                    operandes.append({token: [left, right]})
                last_operator = token
            else:
                raise VQLSyntaxError(f"Unexpected token: {token}")
        assert len(operandes) == 1, operandes
        return operandes[0]

    # print('STACK:', stack)
    infix = expand(stack)
    # print('INFIX:', infix)
    postfix = tuple(as_postfix(infix))
    # print('POSTFIX:', postfix)
    return as_tree(postfix)



from PySide2.QtWidgets import *
from PySide2.QtCore import *
from PySide2.QtGui import *
from enum import Enum

from .abstractquerywidget import AbstractQueryWidget
from cutevariant.core import Query


class FilterType(Enum):
    LOGIC = 1
    CONDITION = 2


class FilterItem(QStandardItem):
    def __init__(self, type=FilterType.LOGIC):
        super().__init__()
        self.type = type

        if self.type == FilterType.LOGIC:
            self.name = "AND"

        else:
            self.name = "pos"
            self.operator = ">"
            self.value = 123

    def makeText(self):
        if self.type == FilterType.LOGIC:
            self.setText(self.name)

        if self.type == FilterType.CONDITION:
            self.setText(self.name + self.operator + self.value)


class FilterQueryModel(QStandardItemModel):
    def __init__(self):
        super().__init__()
        self.query = None

    def setQuery(self, query: Query):
        self.query = query
        self.clear()
        if self.query.filter is not None:
            self.appendRow(self.toItem(self.query.filter))

    def getQuery(self) -> Query:
        if self.rowCount() != 0:
            self.query.filter = self.fromItem(self.item(0))
        return self.query

    def toItem(self, data: dict) -> QStandardItem:

        if len(data) == 1:  #  logic item
            operator = list(data.keys())[0]
            item = FilterItem(FilterType.LOGIC)
            item.name = operator
            item.setEditable(False)
            item.makeText()
            for k in data[operator]:
                item.appendRow(self.toItem(k))
            return item
        else:  # condition item
            item = FilterItem(FilterType.CONDITION)
            item.name = str(data["field"])
            item.operator = str(data["operator"])
            item.value = str(data["value"])
            item.makeText()
            return item

    def fromItem(self, item: QStandardItem) -> dict:
        if item.type == FilterType.LOGIC:
            op = item.name
            data = {op: []}

            for i in range(item.rowCount()):
                data[op].append(self.fromItem(item.child(i)))
            return data

        else:
            data = {}
            data["field"] = item.name
            data["operator"] = item.operator
            data["value"] = item.value
            return data


class FilterEditDialog(QDialog):
    def __init__(self, item: QStandardItem):
        super().__init__()
        self.field = QLineEdit()
        self.operator = QLineEdit()
        self.value = QLineEdit()
        self.item = item

        form_layout = QFormLayout()

        form_layout.addRow("field", self.field)
        form_layout.addRow("operator", self.operator)
        form_layout.addRow("value", self.value)

        layout = QVBoxLayout()
        buttonBox = QDialogButtonBox(QDialogButtonBox.Save | QDialogButtonBox.Cancel)
        buttonBox.accepted.connect(self.save)
        buttonBox.rejected.connect(self.close)

        layout.addLayout(form_layout)
        layout.addWidget(buttonBox)
        self.setLayout(layout)

        self.load()

    def load(self):
        self.field.setText(self.item.data(FilterQueryModel.FieldRole))
        self.operator.setText(self.item.data(FilterQueryModel.OperatorRole))
        self.value.setText(self.item.data(FilterQueryModel.ValueRole))

    def save(self):
        self.item.setText(self.field.text() + self.operator.text() + self.value.text())
        self.item.setData("field", FilterQueryModel.TypeRole)
        self.item.setData(self.field.text(), FilterQueryModel.FieldRole)
        self.item.setData(self.operator.text(), FilterQueryModel.OperatorRole)
        self.item.setData(self.value.text(), FilterQueryModel.ValueRole)


class FilterQueryWidget(AbstractQueryWidget):
    def __init__(self):
        super().__init__()
        self.setWindowTitle("Filter")
        self.view = QTreeView()
        self.model = FilterQueryModel()
        self.view.setModel(self.model)

        layout = QVBoxLayout()
        layout.addWidget(self.view)
        layout.setContentsMargins(0, 0, 0, 0)

        self.setLayout(layout)

        self.model.itemChanged.connect(self.changed)
        self.view.doubleClicked.connect(self.edit)

    def setQuery(self, query: Query):
        self.model.setQuery(query)

    def getQuery(self) -> Query:
        return self.model.getQuery()

    def edit(self, index):
        dialog = FilterEditDialog(self.model.itemFromIndex(index))

        if dialog.exec_():
            dialog.save()

    def contextMenuEvent(self, event):
        """ override methode """

        menu = QMenu(self)

        pos = self.view.viewport().mapFromGlobal(event.globalPos())
        index = self.view.indexAt(pos)

        if index.isValid() is False:
            logic_action = menu.addAction("add logic")

        else:
            logic_action = menu.addAction("add logic")
            item_action = menu.addAction("add condition")
            menu.addSeparator()
            rem_action = menu.addAction("Remove")

        menu.exec_(event.globalPos())



import sqlite3
import sys
import collections


def drop_table(conn, table_name):
    c = conn.cursor()
    c.execute(f"DROP TABLE IF EXISTS {table_name}")


def create_project(conn, name, reference):
    cursor = conn.cursor()
    cursor.execute("""CREATE TABLE projects (name text, reference text NULL)""")
    cursor.execute(
        """INSERT INTO projects VALUES (:name,:reference)""",
        {"name": name, "reference": reference},
    )
    conn.commit()


## ================ SELECTION functions =============================


def create_table_selections(conn):
    """ 
    create selection table and selection_has_variant 

    :param conn: sqlite3.connect
    """

    cursor = conn.cursor()
    cursor.execute(
        """CREATE TABLE selections (name text, count INTEGER NULL, query text NULL )"""
    )
    cursor.execute(
        """CREATE TABLE selection_has_variant (variant_id integer, selection_id integer)"""
    )
    conn.commit()


def insert_selection(conn, name="no_name", count=0, query=str()):
    """ 
    insert one selection

    :param conn: sqlite3.connect
     """
    cursor = conn.cursor()
    cursor.execute(
        """INSERT INTO selections VALUES (:name,:count,:query)""",
        {"name": name, "count": count, "query": query},
    )
    conn.commit()
    return cursor.lastrowid


def create_selection_from_sql(conn, name, query, by="site"):

    #  Create selection
    selection_id = insert_selection(conn, name=name, count=0, query=query)

    # insert into selection_has_variants
    if by == "site":
        q = f"""
        INSERT INTO selection_has_variant 
        SELECT variants.rowid, {selection_id} FROM variants
        INNER JOIN ({query}) query WHERE variants.chr = query.chr AND variants.pos = query.pos
        """

    if by == "variant":
        q = f"""
        INSERT INTO selection_has_variant 
        SELECT variants.rowid, {selection_id} FROM variants 
        INNER JOIN ({query}) as query 
        WHERE variants.chr = query.chr AND variants.pos = query.pos AND variants.ref = query.ref AND variants.alt = query.alt
        """

    cursor = conn.cursor()
    cursor.execute(q)
    conn.commit()

    #  update selection count
    cursor.execute(
        f"""
        UPDATE selections set count = (SELECT COUNT(*) FROM selection_has_variant WHERE selection_id = {selection_id}) WHERE selections.rowid = {selection_id}
        """
    )

    conn.commit()


def get_selections(conn):
    cursor = conn.cursor()
    for row in cursor.execute("""SELECT * FROM selections """):
        record = dict()
        record["name"] = row[0]
        record["count"] = row[1]
        record["query"] = row[2]
        yield record


def intersect_variants(query1, query2, by="site"):

    if by == "site":
        col = "chr,pos"

    if by == "variant":
        col = "chr,pos,ref,alt"

    return f"""
    SELECT {col} FROM ({query1})
    INTERSECT 
    SELECT {col} FROM ({query2})
    """

    return query


def union_variants(query1, query2, by="site"):
    if by == "site":
        col = "chr,pos"

    if by == "variant":
        col = "chr,pos,ref,alt"

    return f"""
    SELECT {col} FROM ({query1})
    UNION
    SELECT {col} FROM ({query2})
    """


def subtract_variants(query1, query2, by="site"):
    if by == "site":
        col = "chr,pos"

    if by == "variant":
        col = "chr,pos,ref,alt"

    return f"""
    SELECT {col} FROM ({query1})
    EXCEPT
    SELECT {col} FROM ({query2})
    """


## ================ Fields functions =============================


def create_table_fields(conn):
    """ 
    create field table 

    :param conn: sqlite3.connect

    """
    cursor = conn.cursor()
    cursor.execute(
        """CREATE TABLE fields
        (name text, category text NULL, type text NULL, description text NULL )
        """
    )
    conn.commit()


def insert_field(
    conn, name="no_name", category="variants", type="text", description=str()
):
    """ 
    insert one field 

    :param conn: sqlite3.connect
    :param name: field name
    :param category: category field name. The default is "variants". Don't use sample as category name
    :param type: sqlite type which can be : integer, real, text
    """
    cursor = conn.cursor()
    cursor.execute(
        """
        INSERT INTO fields VALUES (:name,:category, :type, :description)
        """,
        {"name": name, "category": category, "type": type, "description": description},
    )
    conn.commit()
    return cursor.lastrowid


def insert_many_fields(conn, data: list):
    """ 
    insert many fields using one commit 

    :param conn: sqlite3.connect
    :param data: list of field dictionnary

    :Exemple: 

    insert_many_field(conn, [{name:"sacha", category:"variant", count: 0, description="a description"}])
    insert_many_field(conn, reader.get_fields())

    .. seealso:: insert_field, abstractreader

    """
    cursor = conn.cursor()
    cursor.executemany(
        """
        INSERT INTO fields (name,category,type,description) 
        VALUES (:name,:category,:type, :description)
        """,
        data,
    )
    conn.commit()


def get_fields(conn):
    """ 
    return fields as list of dictionnary 

    :param conn: sqlite3.connect
    :return: list of dictionnary 

    .. seealso:: insert_many_field

    """
    cursor = conn.cursor()

    for row in cursor.execute("""SELECT * FROM fields """):
        record = dict()
        record["name"] = row[0]
        record["category"] = row[1]
        record["type"] = row[2]
        record["description"] = row[3]
        yield record


## ================ Fields functions =============================


def create_table_variants(conn, fields):
    """
    Create variant table which contains dynamics fields 

    :param conn: sqlite3.connect
    :param fields: list of field dictionnary. 
    
    :Example: 
    
    fields = get_fields()
    create_variant_table(conn, fields)

    .. seealso:: get_fields

    """
    cursor = conn.cursor()

    variant_shema = ",".join(
        [
            f'{field["name"]} {field["type"]} NULL'
            for field in fields
            if field["category"] != "sample"
        ]
    )
    cursor.execute(f"""CREATE TABLE variants ({variant_shema})""")
    cursor.execute(
        f"""CREATE TABLE sample_has_variant (sample_id INTEGER, variant_id INTEGER, gt INTEGER DEFAULT -1 )"""
    )

    # cursor.execute(f"""CREATE INDEX sample_has_variant_ids ON sample_has_variant (variant_id, sample_id)""")

    conn.commit()


def get_one_variant(conn, id: int):

    variant = {"rowid": id}

    cols = [i[0] for i in conn.execute("SELECT * FROM variants LIMIT 1").description]
    values = conn.execute(f""" SELECT * FROM variants WHERE rowid = {id}""").fetchone()
    variant.update(dict(zip(cols, values)))
    return variant

    # cursor = conn.cursor()
    # return cursor.fetchone()


def insert_many_variants(conn, data, total_variant_count=-1, commit_every=200):
    """
    Insert many variant from data into variant table.columns

    :param conn: sqlite3.connect
    :param data: list of variant dictionnary which contains same number of key than fields numbers. 
    :param variant_count: total variant count, to compute progression

    :Example: 

    insert_many_variant(conn, [{chr:"chr1", pos:24234, alt:"A","ref":T }]) 
    insert_many_variant(conn, reader.get_variants())
    
    .. warning:: Using reader, this can take a while
    ... todo:: with large dataset, need to cache import   
    .. seealso:: abstractreader
    """

    cursor = conn.cursor()

    #  Get columns description from variant table
    cols = [i[0] for i in conn.execute("SELECT * FROM variants LIMIT 1").description]

    # build dynamic insert query
    # INSERT INTO variant qcol1, qcol2.... VALUES :qcol1, :qcol2 ....
    q_cols = ",".join(cols)
    q_place = ",".join([f":{place}" for place in cols])

    # get samples with sql rowid
    samples = dict(
        [
            (record[1], record[0])
            for record in conn.execute("""SELECT rowid, name FROM samples""")
        ]
    )

    # Loop over variants
    variant_count = 0  # count variants
    insert_count = (
        0
    )  #  count insertion in sql ( one variant can have multiple insertion depending on annotation)

    for variant in data:

        # use default dict for missing value
        variant = collections.defaultdict(lambda: "", variant)

        variant_count += 1

        ## Split variant into multiple variant if there are multiple annotation
        variants = []
        if "annotation" in variant:
            for annotation in variant["annotation"]:
                new_variant = dict(variant)
                new_variant.update(annotation)
                del new_variant["annotation"]
                variants.append(new_variant)
        else:
            variants.append(variant)

        for variant_to_insert in variants:

            # Insert current variant
            cursor.execute(
                f"""INSERT INTO variants ({q_cols}) VALUES ({q_place})""",
                variant_to_insert,
            )

            # get variant rowid
            variant_id = cursor.lastrowid
            insert_count += 1

            #  every commit_every = 200 insert, start a commit ! This value can be changed
            if insert_count % commit_every == 0:
                progress = float(variant_count) / total_variant_count * 100.0
                yield progress, f"{variant_count} variant inserted"
                conn.commit()

            # if variant has sample data, insert record into sample_has_variant
            if "samples" in variant:
                for sample in variant["samples"]:
                    name = sample["name"]
                    gt = sample["gt"]

                    if name in samples.keys():
                        sample_id = samples[name]
                        cursor.execute(
                            f"""INSERT INTO sample_has_variant VALUES (?,?,?)""",
                            [sample_id, variant_id, gt],
                        )

    conn.commit()
    #  create index to make sample query faster
    cursor.execute(
        f"""CREATE UNIQUE INDEX idx_sample_has_variant ON sample_has_variant (sample_id,variant_id)"""
    )

    # create selections
    insert_selection(conn, name="all", count=variant_count)

    yield 100, f"{variant_count} variant(s) has been inserted"


## ================ Fields functions =============================


def create_table_samples(conn):
    """
    Create sample table 

    :param conn: sqlite3.connect

    """
    cursor = conn.cursor()
    cursor.execute("""CREATE TABLE samples (name text)""")
    conn.commit()


def insert_sample(conn, name="no_name"):
    """
    Insert one sample in sample table 

    :param conn: sqlite3.connect

    """
    cursor = conn.cursor()
    cursor.execute(""" INSERT INTO samples VALUES (:name) """, {"name": name})

    conn.commit()
    return cursor.lastrowid


def get_samples(conn):
    """"
    Get samples from sample table 

    :param con: sqlite3.conn 
    :return sample list
    """
    cursor = conn.cursor()
    record = dict()
    for row in cursor.execute("""SELECT name FROM samples"""):
        record["name"] = row[0]
        yield record



from .abstractreader import AbstractReader
import vcf


SNPEFF_ANNOTATION_DEFAULT_FIELDS = {
    "annotation": {
        "name": "consequence",
        "category": "annotation",
        "description": "consequence",
        "type": "text",
    },
    "annotation_impact": {
        "name": "impact",
        "category": "annotation",
        "description": "impact of variant",
        "type": "text",
    },
    "gene_name": {
        "name": "gene",
        "category": "annotation",
        "description": "gene name",
        "type": "text",
    },
    "gene_id": {
        "name": "gene_id",
        "category": "annotation",
        "description": "gene name",
        "type": "text",
    },
    "feature_id": {
        "name": "transcript",
        "category": "annotation",
        "description": "transcript name",
        "type": "text",
    },
    "transcript_biotype": {
        "name": "biotype",
        "category": "annotation",
        "description": " biotype",
        "type": "text",
    },
    "hgvs.p": {
        "name": "hgvs_p",
        "category": "annotation",
        "description": "protein hgvs",
        "type": "text",
    },
    "hgvs.c": {
        "name": "hgvs_c",
        "category": "annotation",
        "description": "coding hgvs",
        "type": "text",
    },
}


class AnnotationParser(object):
    def parse_fields(self, raw):
        self.fields_index = {}  ## required for parse_variant
        for index, field in enumerate(raw.split("|")):
            key = field.strip().lower()

            if key in SNPEFF_ANNOTATION_DEFAULT_FIELDS.keys():
                self.fields_index[index] = SNPEFF_ANNOTATION_DEFAULT_FIELDS[key]["name"]
                yield SNPEFF_ANNOTATION_DEFAULT_FIELDS[key]

    def parse_variant(self, raw):

        annotation = {}
        for index, ann in enumerate(raw.split("|")):
            if index in self.fields_index:
                field_name = self.fields_index[index]
                annotation[field_name] = ann

        return annotation


class VcfReader(AbstractReader):

    type_mapping = {
        "Float": "Float",
        "Integer": "Integer",
        "Flag": "Boolean",
        "String": "String",
    }

    def __init__(self, device):
        super(VcfReader, self).__init__(device)
        self.parser = AnnotationParser()

    def parse_variants(self):
        fields = list(self.parse_fields())
        self.device.seek(0)

        vcf_reader = vcf.Reader(self.device)

        for record in vcf_reader:

            for index, alt in enumerate(record.ALT):
                variant = {
                    "chr": record.CHROM,
                    "pos": record.POS,
                    "ref": record.REF,
                    "alt": str(alt),
                }

                # Read annotations
                for field in fields:
                    category = field["category"]
                    name = field["name"]
                    ftype = field["type"]
                    colname = name
                    value = None

                    # PARSE INFO
                    if category == "info":
                        # Test flags
                        if ftype == "Flag":
                            variant[colname] = True if name in record.INFO else False
                        else:
                            if name in record.INFO:
                                if isinstance(record.INFO[name], list):
                                    value = record.INFO[name][0]
                                else:
                                    value = record.INFO[name]
                            variant[colname] = value

                    # PARSE GENOTYPE / SAMPLE
                    if category == "sample":
                        variant["samples"] = list()
                        for sample in record.samples:
                            gt = -1
                            if sample["GT"] == "0/1":
                                gt = 1
                            if sample["GT"] == "0/0":
                                gt = 0
                            if sample["GT"] == "1/1":
                                gt = 2

                            variant["samples"].append({"name": sample.sample, "gt": gt})

                    #  PARSE Annotation
                    if category == "annotation":
                        # each variant can have multiple annotation. Create then many variants
                        variant["annotation"] = []
                        annotations = record.INFO["ANN"]
                        for annotation in annotations:
                            variant["annotation"].append(
                                self.parser.parse_variant(annotation)
                            )

                yield variant

    def parse_fields(self):

        yield {
            "name": "chr",
            "category": "variant",
            "description": "chromosom",
            "type": "text",
        }
        yield {
            "name": "pos",
            "category": "variant",
            "description": "chromosom",
            "type": "integer",
        }
        yield {
            "name": "ref",
            "category": "variant",
            "description": "chromosom",
            "type": "text",
        }
        yield {
            "name": "alt",
            "category": "variant",
            "description": "chromosom",
            "type": "text",
        }

        self.device.seek(0)
        vcf_reader = vcf.Reader(self.device)
        # Annotation ...
        for key, info in vcf_reader.infos.items():
            if key == "ANN":
                yield from self.parser.parse_fields(info.desc)

        # PEUVENT SE METTRE AUTOMATIQUEMENT ...
        for sample in vcf_reader.samples:
            yield {
                "name": f"gt{sample}.gt",
                "category": "sample",
                "description": "sample genotype",
                "type": "text",
            }

    def get_samples(self):
        self.device.seek(0)
        vcf_reader = vcf.Reader(self.device)
        return vcf_reader.samples



from PySide2.QtWidgets import *
from PySide2.QtCore import *
from PySide2.QtGui import *


from .abstractquerywidget import AbstractQueryWidget
from cutevariant.core import Query
from cutevariant.core import sql

import re


class VqlSyntaxHighlighter(QSyntaxHighlighter):
    def __init__(self, document=None):
        super().__init__(document)

    def highlightBlock(self, text):
        """override """

        palette = qApp.palette("QTextEdit")

        REGEXES = {
            "SELECT|FROM|WHERE": (QFont.Bold, palette.color(QPalette.HighlightedText))
        }
        for regex, (style, color) in REGEXES.items():
            rule = QRegularExpression(regex)
            matchIterator = rule.globalMatch(text)
            while matchIterator.hasNext():
                match = matchIterator.next()
                t_format = QTextCharFormat()
                t_format.setFontWeight(style or QFont.Normal)
                t_format.setForeground(color or Qt.black)
                self.setFormat(match.capturedStart(), match.capturedLength(), t_format)


class VqlEditor(AbstractQueryWidget):
    def __init__(self):
        super().__init__()
        self.setWindowTitle("Columns")
        self.text_edit = QTextEdit()
        self.highlighter = VqlSyntaxHighlighter(self.text_edit.document())

        main_layout = QVBoxLayout()

        main_layout.addWidget(self.text_edit)
        self.setLayout(main_layout)

        self.text_edit.textChanged.connect(self.changed)

    def setQuery(self, query: Query):
        """ Method override from AbstractQueryWidget"""
        self.query = query
        self.text_edit.setPlainText(self.query.to_vql())

    def getQuery(self):
        """ Method override from AbstractQueryWidget"""
        query = self.query.from_vql(self.text_edit.toPlainText())

        return self.query



from PySide2.QtWidgets import *
from PySide2.QtCore import *
from cutevariant.core import Query


class AbstractQueryWidget(QWidget):
    """
    This is a base class for all widget which observe a Query

    .. seealso:: queryrooter.py
    """

    #  signals
    changed = Signal()

    def __init__(self, parent=None):
        super().__init__(parent)

    def setQuery(self, query: Query):
        raise NotImplemented()

    def getQuery(self):
        raise NotImplemented()



from PySide2.QtCore import *
from PySide2.QtWidgets import *
from PySide2.QtGui import *


class VariantPlugin(QWidget):
    def __init__(self):
        super().__init__()

    def set_variant(self, variant):
        raise NotImplemented()



from PySide2.QtWidgets import *
from PySide2.QtCore import *
from PySide2.QtGui import *


from .abstractquerywidget import AbstractQueryWidget
from cutevariant.core import Query
from cutevariant.core import sql


class ColumnQueryModel(QStandardItemModel):
    def __init__(self):
        super().__init__()
        self.setColumnCount(2)
        self.query = None

    def setQuery(self, query: Query):
        self.query = query
        self.load()

    def getQuery(self):
        selected_columns = []
        for i in range(self.rowCount()):
            if self.item(i).checkState() == Qt.Checked:
                selected_columns.append(self.item(i).text())

        self.query.columns = selected_columns

        return self.query

    def load(self):
        self.clear()
        for record in sql.get_fields(self.query.conn):
            item = QStandardItem(record["name"])
            item.setCheckable(True)

            if record["name"] in self.query.columns:
                item.setCheckState(Qt.Checked)

            self.appendRow(item)


class ColumnQueryWidget(AbstractQueryWidget):
    def __init__(self):
        super().__init__()

        self.setWindowTitle("Columns")
        self.view = QTreeView()
        self.model = ColumnQueryModel()
        self.view.setModel(self.model)
        self.view.setIndentation(0)
        self.view.header().setVisible(False)
        layout = QVBoxLayout()

        layout.addWidget(self.view)
        layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(layout)
        self.model.itemChanged.connect(self.changed)

    def setQuery(self, query: Query):
        """ Method override from AbstractQueryWidget"""
        self.model.setQuery(query)

    def getQuery(self):
        """ Method override from AbstractQueryWidget"""
        return self.model.getQuery()



from PySide2.QtWidgets import *
from PySide2.QtCore import *
from PySide2.QtGui import *


from .abstractquerywidget import AbstractQueryWidget
from cutevariant.core import Query
from cutevariant.core import sql


IMPACT_COLOR = {
    "LOW": "#71E096",
    "MODERATE": "#F5A26F",
    "HIGH": "#ed6d79",
    "MODIFIER": "#55abe1",
}


class QueryModel(QAbstractItemModel):
    def __init__(self, parent=None):
        super().__init__()
        self.limit = 50
        self.page = 0
        self.total = 0
        self.query = None
        self.variants = []

    def rowCount(self, parent=QModelIndex()):
        """override"""
        if parent == QModelIndex():
            return len(self.variants)
        return 0

    def columnCount(self, parent=QModelIndex()):
        """override """
        if self.query is None:
            return 0
        else:
            return len(self.query.columns)

    def index(self, row, column, parent):
        """override"""
        if not self.hasIndex(row, column, parent):
            return QModelIndex()

        return self.createIndex(row, column)

    def parent(self, child):
        """ override """
        if not child.isValid():
            return QModelIndex()

        return QModelIndex()

    def data(self, index, role=Qt.DisplayRole):
        """ override """

        if not index.isValid():
            return None

        if role == Qt.DisplayRole:
            # First row is variant id  don't show
            return str(self.variants[index.row()][index.column() + 1])

        return None

    def headerData(self, section, orientation, role=Qt.DisplayRole):
        """override"""
        if orientation == Qt.Horizontal:
            if role == Qt.DisplayRole:
                return self.query.columns[section]

        if orientation == Qt.Vertical:
            if role == Qt.DisplayRole:
                return self.variants[index.row()][0]

        return None

    def setQuery(self, query: Query):
        self.query = query
        self.total = query.count()
        self.load()

    def load(self):
        self.beginResetModel()
        self.variants.clear()
        self.variants = list(self.query.rows(self.limit, self.page * self.limit))
        self.endResetModel()

    def hasPage(self, page):
        return page >= 0 and page * self.limit < self.total

    def setPage(self, page):
        if self.hasPage(page):
            self.page = page
            self.load()

    def nextPage(self):
        if self.hasPage(self.page + 1):
            self.setPage(self.page + 1)

    def previousPage(self):
        if self.hasPage(self.page - 1):
            self.setPage(self.page - 1)

    def sort(self, column: int, order):
        """override"""
        pass
        if column < self.columnCount():
            colname = self.query.columns[column]

            print("ORDER", order)
            self.query.order_by = colname
            self.query.order_desc = True if order == Qt.DescendingOrder else False
            self.load()

    def get_rowid(self, index):
        return self.variants[index.row()][0]


class QueryDelegate(QStyledItemDelegate):
    def paint(self, painter, option, index):
        """overriden"""

        palette = qApp.palette("QTreeView")
        colname = index.model().headerData(index.column(), Qt.Horizontal)
        value = index.data(Qt.DisplayRole)
        select = option.state & QStyle.State_Selected

        #  Draw selection background
        if select:
            bg_brush = palette.brush(QPalette.Highlight)
        #  Draw alternative
        else:
            if index.row() % 2:
                bg_brush = palette.brush(QPalette.Midlight)
            else:
                bg_brush = palette.brush(QPalette.Light)

        painter.setBrush(bg_brush)
        painter.setPen(Qt.NoPen)
        painter.drawRect(option.rect)

        if colname == "impact":
            painter.setPen(QPen(IMPACT_COLOR.get(value, palette.color(QPalette.Text))))
            painter.drawText(option.rect, Qt.AlignLeft | Qt.AlignVCenter, index.data())
            return

        painter.setPen(
            QPen(palette.color(QPalette.HighlightedText if select else QPalette.Text))
        )
        painter.drawText(option.rect, Qt.AlignLeft | Qt.AlignVCenter, index.data())

    def draw_biotype(self, value):
        pass

    def sizeHint(self, option, index):
        """override"""
        return QSize(0, 30)


class ViewQueryWidget(AbstractQueryWidget):

    variant_clicked = Signal(dict)

    def __init__(self):
        super().__init__()
        self.model = QueryModel()
        self.delegate = QueryDelegate()
        # self.delegate = VariantDelegate()
        self.setWindowTitle("Variants")
        self.topbar = QToolBar()
        self.bottombar = QToolBar()
        self.view = QTreeView()

        self.view.setFrameStyle(QFrame.NoFrame)
        self.view.setModel(self.model)
        self.view.setItemDelegate(self.delegate)
        self.view.setAlternatingRowColors(True)
        self.view.setSortingEnabled(True)
        # self.view.setItemDelegate(self.delegate)

        main_layout = QVBoxLayout()
        main_layout.addWidget(self.topbar)
        main_layout.addWidget(self.view)
        main_layout.addWidget(self.bottombar)
        main_layout.setContentsMargins(0, 0, 0, 0)

        # Construct top bar
        self.topbar.addAction("save")

        # Construct bottom bar
        self.page_info = QLabel()
        self.page_box = QLineEdit()
        self.page_box.setReadOnly(True)
        self.page_box.setFrame(QFrame.NoFrame)
        self.page_box.setMaximumWidth(50)
        self.page_box.setAlignment(Qt.AlignHCenter)
        self.page_box.setStyleSheet("QWidget{background-color: transparent;}")
        self.page_box.setText("0")
        spacer = QWidget()
        spacer.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Preferred)
        self.bottombar.addAction("sql")
        self.bottombar.addWidget(self.page_info)
        self.bottombar.addWidget(spacer)
        self.bottombar.addAction("<", self.model.previousPage)
        self.bottombar.addWidget(self.page_box)
        self.bottombar.addAction(">", self.model.nextPage)

        self.bottombar.setContentsMargins(0, 0, 0, 0)

        self.setLayout(main_layout)

        self.model.modelReset.connect(self.updateInfo)

        # emit variant when clicked
        self.view.clicked.connect(self._variant_clicked)

    def setQuery(self, query: Query):
        """ Method override from AbstractQueryWidget"""
        self.model.setQuery(query)

    def getQuery(self):
        """ Method override from AbstractQueryWidget"""
        return self.model.query

    def updateInfo(self):

        self.page_info.setText(f"{self.model.total} variant(s)")
        self.page_box.setText(f"{self.model.page}")

    def _variant_clicked(self, index):
        print("cicked on ", index)
        rowid = self.model.get_rowid(index)
        variant = sql.get_one_variant(self.model.query.conn, rowid)
        self.variant_clicked.emit(variant)



from PySide2.QtWidgets import *
from PySide2.QtCore import *
import sys
from cutevariant.gui import MainWindow


def main():
    """The main routine."""

    app = QApplication(sys.argv)
    w = MainWindow()
    w.show()
    app.exec_()


if __name__ == "__main__":

    main()



from PySide2.QtWidgets import *
from PySide2.QtCore import *

import os
import sqlite3
from cutevariant.core.importer import async_import_file


class ProjetPage(QWizardPage):
    def __init__(self):
        super().__init__()

        self.setTitle("Project creation")
        self.setSubTitle("This wizard will guide you to create a cutevariant project.")

        self.projet_name_edit = QLineEdit()
        self.projet_path_edit = QLineEdit()
        self.projet_path_edit.setText(os.getcwd())
        self.browse_button = QPushButton("Browse")
        self.reference = QComboBox()

        self.reference.addItem("hg19")
        self.registerField("project_name", self.projet_name_edit, "text")
        self.registerField("project_path", self.projet_path_edit, "text")
        self.registerField("reference", self.reference, "currentText")

        v_layout = QFormLayout()

        browse_layout = QHBoxLayout()
        browse_layout.addWidget(self.projet_path_edit)
        browse_layout.addWidget(self.browse_button)
        browse_layout.setContentsMargins(0, 0, 0, 0)

        v_layout.addRow("Reference genom", self.reference)
        v_layout.addRow("Project Name", self.projet_name_edit)
        v_layout.addRow("Create in", browse_layout)

        self.setLayout(v_layout)

        self.browse_button.clicked.connect(self._browse)
        self.projet_path_edit.textChanged.connect(self.completeChanged)
        self.projet_name_edit.textChanged.connect(self.completeChanged)

    def _browse(self):
        path = QFileDialog.getExistingDirectory(self, "Select a path for the project")
        if path:
            self.projet_path_edit.setText(path)

    def isComplete(self):
        """Conditions to unlock next button"""
        return True if (
            QDir(self.projet_path_edit.text()).exists()
            and self.projet_path_edit.text()
            and self.projet_name_edit.text()
        ) else False


class FilePage(QWizardPage):
    def __init__(self):
        super().__init__()

        self.setTitle("Select a file")
        self.setSubTitle("Supported file are vcf and vcf.gz.")

        self.file_path_edit = QLineEdit()
        self.button = QPushButton("Browse")
        v_layout = QHBoxLayout()

        v_layout.addWidget(self.file_path_edit)
        v_layout.addWidget(self.button)
        self.setLayout(v_layout)

        self.button.clicked.connect(self._browse)
        self.file_path_edit.textChanged.connect(self.completeChanged)

        self.registerField("filename", self.file_path_edit, "text")

    def _browse(self):
        filename = QFileDialog.getOpenFileName(
            self, "Open a file", QDir.homePath(), "VCF file (*.vcf, *.vcf.gz)"
        )
        if filename:
            self.file_path_edit.setText(filename[0])

    def isComplete(self):
        """Conditions to unlock next button"""
        return True if (
            self.file_path_edit.text() and QFile(self.file_path_edit.text()).exists()
        ) else False


class ImportThread(QThread):

    progress_changed = Signal(int, str)
    finished_status = Signal(bool)

    def __init__(self):
        super().__init__()
        self._stop = False
        self.filename = None
        self.db_filename = None

    def run(self):
        """Overrided QThread method

        .. warning:: This method is the only one to be executed in the thread.
            SQLite objects created in a thread can only be used in that same
            thread; so no use of self.conn is allowed elsewhere.
        """
        self._stop = False

        if self.db_filename is None:
            self.db_filename = self.filename + ".db"

        if os.path.exists(self.db_filename):
            os.remove(self.db_filename)
        self.conn = sqlite3.connect(self.db_filename)

        for value, message in async_import_file(self.conn, self.filename):
            if self._stop == True:
                self.conn.close()
                break
            # Send progression
            self.progress_changed.emit(value, message)
        # Send status (True when there is no error)
        self.finished_status.emit(not self._stop)

    def stop(self):
        self._stop = True


class ImportPage(QWizardPage):
    def __init__(self):
        super().__init__()

        self.setTitle("Import file")
        self.setSubTitle("Please click on Import/Stop to start or stop the process.")
        self.text_buttons = ["Import", "Stop"]

        self.thread_finished = False
        self.thread = ImportThread()
        self.progress = QProgressBar()
        self.import_button = QPushButton(self.text_buttons[0])

        v_layout = QHBoxLayout()
        v_layout.addWidget(self.progress)
        v_layout.addWidget(self.import_button)

        m_layout = QVBoxLayout()

        self.log_edit = QPlainTextEdit()
        self.tab_widget = QTabWidget()
        self.tab_widget.addTab(self.log_edit, "log")

        m_layout.addLayout(v_layout)
        m_layout.addWidget(self.tab_widget)
        self.setLayout(m_layout)

        self.log_edit.appendPlainText(self.field("filename"))

        self.thread.started.connect(lambda: self.log_edit.appendPlainText("Started"))
        self.thread.finished.connect(lambda: self.log_edit.appendPlainText("Done"))
        self.import_button.clicked.connect(self.run)

        self.thread.progress_changed.connect(self.progress_changed)
        self.thread.finished.connect(self.import_thread_finished)
        self.thread.finished_status.connect(self.import_thread_finished_status)

    def progress_changed(self, value, message):
        self.progress.setValue(value)
        self.log_edit.appendPlainText(message)

    def import_thread_finished(self):
        """Force the activation of the finish button after a successful import"""
        self.completeChanged.emit()

    def import_thread_finished_status(self, status):
        """Set the finished status of import thread

        .. note:: Called at the end of run()
        """
        self.thread_finished = status

        if status:
            self.import_button.setDisabled(True)

    def run(self):
        if self.thread.isRunning():
            print("stop thread")
            self.import_button.setDisabled(True)
            self.thread.stop()
            self.progress.setValue(0)
            self.import_button.setDisabled(False)
            self.import_button.setText(self.text_buttons[0])

        else:
            self.thread.filename = self.field("filename")
            self.thread.db_filename = (
                self.field("project_path")
                + QDir.separator()
                + self.field("project_name")
                + ".db"
            )

            self.log_edit.appendPlainText("Import " + self.thread.filename)
            self.import_button.setText(self.text_buttons[1])
            self.thread.start()

    def isComplete(self):
        """Conditions to unlock next button"""
        return self.thread_finished


class ProjetWizard(QWizard):
    def __init__(self):
        super().__init__()
        self.setWindowTitle("Cutevariant - Project creation wizard")

        self.addPage(ProjetPage())
        self.addPage(FilePage())
        self.addPage(ImportPage())



from .abstractreader import AbstractReader
import csv


class CsvReader(AbstractReader):
    def __init__(self, device):
        super(CsvReader, self).__init__(device)

    def __del__(self):
        del (self.device)

    def get_samples(self):
        return ["boby", "sacha", "olivier"]

    def parse_fields(self):
        self.device.seek(0)
        csvreader = csv.reader(self.device, delimiter="\t")
        rows = next(csvreader)
        for row in rows:
            row = row.replace("#", "")
            yield {"name": row, "type": "text", "category": None, "description": None}

    def parse_variants(self):
        self.device.seek(0)
        csvreader = csv.reader(self.device, delimiter="\t")
        next(csvreader)
        for row in csvreader:
            variant = {}
            variant["chr"] = row[0]
            variant["pos"] = row[1]
            variant["ref"] = row[2]
            variant["alt"] = row[3]

            # testing purpose
            variant["samples"] = [
                {"name": "boby", "gt": 0},
                {"name": "sacha", "gt": 1},
                {"name": "olivier", "gt": 2},
            ]

            yield variant


if __name__ == "__main__":
    print("yello")



from PySide2.QtCore import *
from PySide2.QtWidgets import *
import sys


class BrowseFileEdit(QWidget):
    def __init__(self, parent=None):
        super(BrowseFileEdit, self).__init__(parent)
        self.edit = QLineEdit()
        self.button = QPushButton("Browse")
        self.main_layout = QHBoxLayout()

        self.main_layout.addWidget(self.edit)
        self.main_layout.addWidget(self.button)
        self.main_layout.setContentsMargins(0, 0, 0, 0)

        self.setLayout(self.main_layout)

        self.button.clicked.connect(self.browse)

    def browse(self):
        filename = QFileDialog.getOpenFileName(self)[0]
        if filename:
            self.edit.setText(filename)


# app = QApplication(sys.argv)


# w = BrowseFileEdit()

# w.show()

# app.exec()



from PySide2.QtWidgets import *
from PySide2.QtCore import *
from PySide2.QtGui import *


from .abstractquerywidget import AbstractQueryWidget
from cutevariant.core import Query
from cutevariant.core import sql


class SelectionQueryModel(QStandardItemModel):
    def __init__(self):
        super().__init__()
        self.setColumnCount(2)

    def setQuery(self, query: Query):
        self.query = query
        self.refresh()

    def refresh(self):
        self.clear()
        for record in sql.get_selections(self.query.conn):
            name_item = QStandardItem(record["name"])
            count_item = QStandardItem(str(record["count"]))
            self.appendRow([name_item, count_item])

    def save_current_query(self, name):
        sql.create_selection_from_sql(self.query.conn, name, self.query.sql())
        self.refresh()


class SelectionQueryWidget(AbstractQueryWidget):
    def __init__(self):
        super().__init__()

        self.setWindowTitle("Selections")
        self.view = QTreeView()
        self.model = SelectionQueryModel()
        self.view.setModel(self.model)

        layout = QVBoxLayout()

        layout.addWidget(self.view)
        layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(layout)

        self.view.selectionModel().currentRowChanged.connect(self.changed)

    def setQuery(self, query: Query):
        """ Method override from AbstractQueryWidget"""
        self.model.setQuery(query)

    def getQuery(self):
        """ Method override from AbstractQueryWidget"""

        item = self.model.item(self.view.selectionModel().currentIndex().row())

        query = self.model.query
        query.selection = item.text()

        return query

    def save_current_query(self):

        name, success = QInputDialog.getText(
            self, "type a name for selection", "Selection name:", QLineEdit.Normal
        )

        if success:
            self.model.save_current_query(name)
